import asyncio
import json
from collections import namedtuple
from typing import List

"""
Callback Listener is a tuple with `event` and `callback` 
"""
CallbackListener = namedtuple("CallbackListener", "event callback")


class Channel:
    """
    `Channel` is an abstraction for a topic listener for an existing socket connection.
    Each Channel has its own topic and a list of event-callbacks that responds to messages.
    Should only be instantiated through `connection.Socket().set_chanel(topic)`
    Topic-Channel has a 1-many relationship.
    """

    def __init__(self, socket, topic: str, params: dict = {}):
        """
        :param socket: Socket object
        :param topic: Topic that it subscribes to on the realtime server
        :param params:
        """
        self.socket = socket
        self.topic: str = topic
        self.params: dict = params
        self.listeners: List[CallbackListener] = []
        self.joined: bool = False

    def join(self):
        """
        Wrapper for async def _join() to expose a non-async interface
        Essentially gets the only event loop and attempt joining a topic
        :return: None
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._join())
        return self

    async def _join(self):
        """
        Coroutine that attempts to join Phoenix Realtime server via a certain topic
        :return: Channel.channel
        """
        join_req = dict(topic=self.topic, event="phx_join", payload={}, ref=None)

        try:
            await self.socket.ws_connection.send(json.dumps(join_req))

        except Exception as e:
            print(str(e))
            return

    def on(self, event: str, callback):
        """
        :param event: A specific event will have a specific callback
        :param callback: Callback that takes msg payload as its first argument
        :return: None
        """

        cl = CallbackListener(event=event, callback=callback)
        self.listeners.append(cl)
        return self

    def off(self, event: str):
        """
        :param event: Stop responding to a certain event
        :return: None
        """
        self.listeners = [callback for callback in self.listeners if callback.event != event]



import asyncio
import json
import logging
from collections import defaultdict
from functools import wraps

import websockets

from realtime_py.channel import Channel
from realtime_py.exceptions import NotConnectedError
from realtime_py.message import HEARTBEAT_PAYLOAD, PHOENIX_CHANNEL, ChannelEvents, Message

logging.basicConfig(format="%(asctime)s:%(levelname)s - %(message)s", level=logging.INFO)


class Socket:
    def ensure_connection(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not args[0].connected:
                raise NotConnectedError(func.__name__)

            return func(*args, **kwargs)

        return wrapper

    def __init__(self, url: str, params: dict = {}, hb_interval: int = 5):
        """
        `Socket` is the abstraction for an actual socket connection that receives and 'reroutes' `Message` according to its `topic` and `event`.
        Socket-Channel has a 1-many relationship.
        Socket-Topic has a 1-many relationship.
        :param url: Websocket URL of the Realtime server. starts with `ws://` or `wss://`
        :param params: Optional parameters for connection.
        :param hb_interval: WS connection is kept alive by sending a heartbeat message. Optional, defaults to 5.
        """
        self.url = url
        self.channels = defaultdict(list)
        self.connected = False
        self.params: dict = params
        self.hb_interval: int = hb_interval
        self.ws_connection: websockets.client.WebSocketClientProtocol = None
        self.kept_alive: bool = False

    @ensure_connection
    def listen(self):
        """
        Wrapper for async def _listen() to expose a non-async interface
        In most cases, this should be the last method executed as it starts an infinite listening loop.
        :return: None
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.gather(self._listen(), self._keep_alive()))

    async def _listen(self):
        """
        An infinite loop that keeps listening.
        :return: None
        """
        while True:
            try:
                msg = await self.ws_connection.recv()
                msg = Message(**json.loads(msg))
                if msg.event == ChannelEvents.reply:
                    continue
                for channel in self.channels.get(msg.topic, []):
                    for cl in channel.listeners:
                        if cl.event == msg.event:
                            cl.callback(msg.payload)

            except websockets.exceptions.ConnectionClosed:
                logging.exception("Connection closed")
                break

    def connect(self):
        """
        Wrapper for async def _connect() to expose a non-async interface
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._connect())
        self.connected = True

    async def _connect(self):

        ws_connection = await websockets.connect(self.url)
        if ws_connection.open:
            logging.info("Connection was successful")
            self.ws_connection = ws_connection
            self.connected = True

        else:
            raise Exception("Connection Failed")

    async def _keep_alive(self):
        """
        Sending heartbeat to server every 5 seconds
        Ping - pong messages to verify connection is alive
        """
        while True:
            try:
                data = dict(
                    topic=PHOENIX_CHANNEL,
                    event=ChannelEvents.heartbeat,
                    payload=HEARTBEAT_PAYLOAD,
                    ref=None,
                )
                await self.ws_connection.send(json.dumps(data))
                await asyncio.sleep(self.hb_interval)
            except websockets.exceptions.ConnectionClosed:
                logging.exception("Connection with server closed")
                break

    @ensure_connection
    def set_channel(self, topic: str):
        """
        :param topic: Initializes a channel and creates a two-way association with the socket
        :return: None
        """

        chan = Channel(self, topic, self.params)
        self.channels[topic].append(chan)

        return chan

    def summary(self):
        """
        Prints a list of topics and event the socket is listening to
        :return: None
        """
        for topic, chans in self.channels.items():
            for chan in chans:
                print(f"Topic: {topic} | Events: {[e for e, _ in chan.callbacks]}]")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jl_realtime_py']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses>=0.6,<0.7',
 'python-dateutil>=2.8.1,<3.0.0',
 'websockets>=8.1,<9.0']

setup_kwargs = {
    'name': 'jl-realtime-py',
    'version': '0.2.1',
    'description': '',
    'long_description': None,
    'author': 'None',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
